/*
 * Copyright 2002, 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_IOCCOM_H
#include <sys/ioccom.h>
#endif
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/tree.h>
#include <sys/wait.h>
#include <sys/queue.h>

#include <pcap.h>

#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#include <unistd.h>
#include <getopt.h>
#include <dnet.h>

#undef timeout_pending
#undef timeout_initialized

#include <event.h>

#include "nttlscan.h"
#include "interface.h"

intf_t *nttl_intf;
struct interface *nttl_inter;
rand_t *nttl_rand;
ip_t   *nttl_ip;
struct addr *nttl_range;		/* destination ip addresses */
struct addr *nttl_srange;		/* source ip addresses */
static int nttl_maxinflight = NTTL_MAXINFLIGHT;
static int nttl_activeprobes;
static int nttl_nprobes;
static int nttl_maxprobes = 100;
static int tcpmode = 0;			/* measure packet rate */
static int (*fnprobe)(struct addr *, struct addr *, int, int) = tcp_probe;

extern int tcprate;

/* Structure for probes */
static SPLAY_HEAD(probetree, probe) probes;

int
probecompare(struct probe *a, struct probe *b)
{
	return (addr_cmp(&a->dst, &b->dst));
}

SPLAY_PROTOTYPE(probetree, probe, node, probecompare);
SPLAY_GENERATE(probetree, probe, node, probecompare);

void
add_probes(struct addr *range)
{
	int nprobes, i;

	/* Do not schedule new probes if we have too many in flight already */
	if (nttl_activeprobes >= nttl_maxinflight)
		return;

	nprobes = nttl_maxprobes - (nttl_nprobes + nttl_activeprobes);
	if (nprobes > nttl_maxinflight - nttl_activeprobes)
		nprobes = nttl_maxinflight - nttl_activeprobes;

	for (i = 0; i < nprobes; i++)
		make_probe(range);
}

void
finalize_probe(struct probe *probe)
{
	int i;

	fprintf(stdout, "%s: ", addr_ntoa(&probe->dst));
	for (i = 0; i < probe->nhops; i++) {
		struct hop *hop = probe->hops[i];

		if ((hop->flags & HOP_STORED) == 0) {
			fprintf(stdout, "* ");
			continue;
		}
		if ((hop->flags & HOP_ENDHOST) == 0) {
			char *special = "";

			if (hop->flags & HOP_UNREACH) {
				switch (hop->unreach) {
				case ICMP_UNREACH_NET:
					special = " (N!)";
					break;
				case ICMP_UNREACH_HOST:
					special = " (N!)";
					break;
				case ICMP_UNREACH_PORT:
					special = " (p!)";
					break;
				case ICMP_UNREACH_HOST_PROHIB:
					special = " (HP!)";
					break;
				case ICMP_UNREACH_FILTER_PROHIB:
					special = " (FP!)";
					break;
				default:
					special = " (U!)";
					break;
				}
				
			}
			fprintf(stdout, "%s%s ",
			    addr_ntoa(&hop->addr), special);
			if (hop->flags & HOP_UNREACH)
				break;
		} else {
			fprintf(stdout, "%s (E)", addr_ntoa(&hop->addr));
			break;
		}
	}
	fprintf(stdout, "\n");

	nttl_activeprobes--;
	nttl_nprobes++;

	if (nttl_activeprobes == 0) {
		fprintf(stderr, "\n");
		nttl_exit(0);
	}

	add_probes(nttl_range);

	fprintf(stderr, "Active: %8d, N: %8d   \r",
	    nttl_activeprobes, nttl_nprobes);
}

void
probe_timeout(int fd, short what, void *arg)
{
	struct hop *hop = arg;
	struct probe *probe = hop->parent;

	/* Reduce active hops */
	probe->active--;

	hop->count++;
	if (hop->count <= PROBE_MAXRETRIES) {
		int res;

		res = fnprobe(&probe->src, &probe->dst, hop->ttl, hop->count);
		/* 
		 * If we encounter a transmit error, then we need to reduce
		 * the count, so that the probe gets repeated.
		 */
		if (res == -1)
			hop->count--;
		schedule_probe(hop, hop->count * hop->count * PROBE_TIMEOUT);
		return;
	}

	hop->flags |= HOP_TIMEOUT;

	/* Check if there are pending events */
	if (probe->active == 0)
		finalize_probe(probe);
}

void
unschedule_probe(struct hop *hop)
{
	if (evtimer_pending(&hop->ev, NULL)) {
		struct probe *probe = hop->parent;

		evtimer_del(&hop->ev);
		probe->active--;
	}
}

void
schedule_probe(struct hop *hop, int ms)
{
	struct probe *probe = hop->parent;
	struct timeval tv;

	evtimer_set(&hop->ev, probe_timeout, hop);

	timerclear(&tv);
	tv.tv_usec = ms * 1000;
	tv.tv_usec += rand_uint16(nttl_rand) & 0x1fff;
	if (tv.tv_usec > 1000000L) {
		tv.tv_sec = tv.tv_usec / 1000000L;
		tv.tv_usec = tv.tv_usec % 1000000L;
	}
	evtimer_add(&hop->ev, &tv);

	probe->active++;
}

void
store_result(struct probe *probe, struct addr *addr,
    u_int ttl, u_int count, int type, int code)
{
	struct hop *hop;
	int endreached = 0;

	/* Bad TTL */
	if (ttl == 0 || ttl > probe->nhops)
		return;

	/* Bad retry count */
	if (count >= PROBE_MAXRETRIES)
		return;

	hop = probe->hops[ttl - 1];

	if ((hop->flags & HOP_STORED) != 0)
		return;
	/* Remove timeout */
	unschedule_probe(hop);

	/* Storing hop has to happen after unschedule */
	hop->flags |= HOP_STORED;

	hop->addr = *addr;

	if (type != DECODE_TCP_TCP) {
		if (type == DECODE_UNREACH) {
			hop->flags |= HOP_UNREACH;
			hop->unreach = code;
			endreached = 1;
		} else {
			/* Time exceeded for last two, increase hops */
			if (probe->nhops - ttl < 2)
				allocate_hops(probe, 10);
		}
	} else {
		hop->flags |= HOP_ENDHOST;
		endreached = 1;
	}

	if (endreached) {
		int i;

		/* End reached, cancel remaining timeouts */
		for (i = ttl; i < probe->nhops; i++) {
			struct hop *tmp = probe->hops[i];

			/* 
			 * If we should receive replies for a probe anyway,
			 * we need to make sure that we ignore them.  This
			 * can happen if we overshot the ttl to a reachable
			 * host.
			 */ 
			tmp->flags |= HOP_STORED;

			unschedule_probe(tmp);
		}
	}

	if (probe->active == 0)
		finalize_probe(probe);
}

/*
 * Allocate more hops to the structure if necessary.
 * The probes are scheduled for network execution.
 */

void
allocate_hops(struct probe *probe, int nr)
{
	int i, ms;

	if (probe->nhops + nr > PROBE_MAXHOPS)
		nr = PROBE_MAXHOPS - probe->nhops;

	if (nr == 0)
		return;

	for (ms = 0, i = probe->nhops; i < probe->nhops + nr; i++, ms++) {
		struct hop *hop;

		if ((hop = calloc(1, sizeof(struct hop))) == NULL)
			err(1, "%s: calloc", __func__);

		evtimer_set(&hop->ev, probe_timeout, hop);
		hop->parent = probe;
		hop->ttl = i + 1;

		probe->hops[i] = hop;

		schedule_probe(hop, ms * PROBE_INITIALDELAY);
	}

	probe->nhops += nr;
}

int
make_probe(struct addr *range)
{
	struct intf_entry intf;
	struct probe *probe;

	if ((probe = calloc(1, sizeof(struct probe))) == NULL)
		err(1, "%s: calloc", __func__);

	do {
		/* Create a random IP adress */
		make_random_ip(range, &probe->dst);
		
		/* Find the IP address of the outgoing interface */
		intf.intf_len = sizeof(intf);
		intf_get_dst(nttl_intf, &intf, &probe->dst);
		probe->src = intf.intf_addr;
	} while (SPLAY_FIND(probetree, &probes, probe) != NULL);

	SPLAY_INSERT(probetree, &probes, probe);

	allocate_hops(probe, 10);

	nttl_activeprobes++;

	return (0);
}

struct probe *
find_probe(struct addr *a)
{
	struct probe tmp;

	tmp.dst = *a;
	return (SPLAY_FIND(probetree, &probes, &tmp));
}

void
nttl_recv_cb(u_char *ag, const struct pcap_pkthdr *pkthdr, const u_char *pkt)
{
	struct interface *inter = (struct interface *)ag;
	struct ip_hdr *ip, *rip;
	struct tcp_hdr *tcp = NULL;
	struct icmp_hdr *icmp = NULL;
	struct icmp_msg_quote *icmp_quote;
	struct addr addr;
	u_short iplen, iphlen, quotelen;
	int which = 0;

	if (nttl_inter == NULL)
		nttl_inter = inter;

	/* Everything below assumes that the packet is IPv4 */
	if (pkthdr->caplen < inter->if_dloff + IP_HDR_LEN)
		return;

	pkt += inter->if_dloff;
	ip = (struct ip_hdr *)pkt;

	iplen = ntohs(ip->ip_len);
	if (pkthdr->caplen - inter->if_dloff < iplen)
		return;

	iphlen = ip->ip_hl << 2;
	if (iphlen > iplen)
		return;
	if (iphlen < sizeof(struct ip_hdr))
		return;

	addr_pack(&addr, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_src, IP_ADDR_LEN);

	switch (ip->ip_p) {
	case IP_PROTO_ICMP:
		/* Ignore ICMP packets in TCP mode */
		if (tcpmode)
			return;

		if (iplen < iphlen + ICMP_HDR_LEN)
			return;

		icmp = (struct icmp_hdr *)(pkt + iphlen);
		switch (icmp->icmp_type) {
		case ICMP_TIMEXCEED:
			if (icmp->icmp_code != ICMP_TIMEXCEED_INTRANS)
				goto bad_icmp_code;
			which = DECODE_EXCEED;
			break;
		case ICMP_UNREACH:
			switch (icmp->icmp_code) {
			case ICMP_UNREACH_NET:
			case ICMP_UNREACH_HOST:
			case ICMP_UNREACH_PORT:
			case ICMP_UNREACH_HOST_PROHIB:
			case ICMP_UNREACH_FILTER_PROHIB:
				break;
			default:
				goto bad_icmp_code;
			}
			which = DECODE_UNREACH;
			break;
		default:
			goto bad_icmp_code;
		}

		quotelen = iplen - iphlen - sizeof(struct icmp_hdr);
		if (quotelen < sizeof(struct ip_hdr))
			goto bad_quotelen;

		icmp_quote = (struct icmp_msg_quote *)(icmp + 1);
		rip = (struct ip_hdr *)(&icmp_quote->icmp_ip);

		switch (rip->ip_p) {
		case IP_PROTO_TCP:
			if (quotelen <
			    (rip->ip_hl << 2) + 12)
				goto bad_quotelen;

			decode_tcp_probe(&addr, rip, which, icmp->icmp_code);
			break;
		case IP_PROTO_UDP:
			if (quotelen <
			    (rip->ip_hl << 2) + 12)
				goto bad_quotelen;

			decode_udp_probe(&addr, rip, which, icmp->icmp_code);
			break;
		default:
			fprintf(stderr, "%s: bad probe protocol: %d\n",
			    __func__, rip->ip_p);
			break;
		}
		break;

	case IP_PROTO_TCP:
		if (iplen < iphlen + TCP_HDR_LEN)
			return;

		tcp = (struct tcp_hdr *)(pkt + iphlen);

		/* Ignore our own packets */
		if ((tcp->th_flags & (TH_SYN|TH_ACK|TH_RST)) == TH_SYN)
			return;

		if (tcpmode)
			decode_tcp_connect(&addr, ip);
		else
			decode_tcp_probe(&addr, ip, DECODE_TCP_TCP, 0);
		break;

	default:
		fprintf(stderr, "Got packet from %s\n", addr_ntoa(&addr));
	}
	return;

 bad_quotelen:
	fprintf(stderr, "%s: received too short ICMP quote (%d) from %s\n",
	    __func__, quotelen, addr_ntoa(&addr));
	return;

 bad_icmp_code:
	fprintf(stderr, "%s: received bad ICMP code from %s: %d %d\n",
	    __func__, addr_ntoa(&addr),
	    icmp->icmp_type, icmp->icmp_code);
	return;

}

void
usage(void)
{
	fprintf(stderr,
		"Usage: nttlscan [OPTIONS] [net ...]\n\n"
		"where options include:\n"
	        "  -s <net>		  Source address range.\n"
	        "  -d <net>		  Destination address range.\n"
		"  -m			  Maximum number of addresses.\n"
		"  -M			  Max number of probes in flight.\n"
	        "  -t			  TCP scan mode.\n"
		"  -h                     Print this message and exit.\n"
	    );
	
	exit(1);
}

void
nttl_exit(int status)
{
	interface_close_all();

	ip_close(nttl_ip);
	rand_close(nttl_rand);

	exit(status);
}

void
nttl_signal(int fd, short what, void *arg)
{
	fprintf(stderr, "exiting on signal %d\n", fd);
	nttl_exit(0);
}

int
addr_inrange(struct addr *net, struct addr *addr)
{
	struct addr addr_a;
	struct addr addr_aend;

	/* Set up the addresses; still IPv4 dependent */
	addr_a = *net;
	addr_a.addr_bits = IP_ADDR_BITS;

	addr_bcast(net, &addr_aend);
	addr_aend.addr_bits = IP_ADDR_BITS;

	if (addr_cmp(&addr_aend, addr) < 0)
		return (0);
	if (addr_cmp(addr, &addr_a) < 0)
		return (0);

	return (1);
}

char *unusednets[] = {
	"127.0.0.0/8",		/* local */
	"10.0.0.0/8",		/* rfc-1918 */
	"172.16.0.0/12",	/* rfc-1918 */
	"192.168.0.0/16",	/* rfc-1918 */
	"224.0.0.0/4",		/* rfc-1112 */
	"240.0.0.0/4",
	"0.0.0.0/8",
	"255.0.0.0/8",
	NULL
};

int
valid_ip(struct addr *addr)
{
	struct addr net;
	char **anet;

	for (anet = &unusednets[0]; *anet != NULL; anet++) {
		addr_aton(*anet, &net);
		if (addr_inrange(&net, addr))
			return (0);
	}

	return (1);
}

/*
 * Creates a random IP address in a certain range.
 * 
 * If the range is NULL, it creates random IP addresses within the
 * routeable address space.
 */

void
make_random_ip(struct addr *range, struct addr *result)
{
	uint32_t rand;
	uint8_t nonce;
	struct addr src;
	int valid = 0, i;

	while (!valid) {
		rand = 0;
		for (i = 0; i < 4; i++) {
			rand *= 256;
			nonce = rand_uint8(nttl_rand);
			if (nonce == 0)
				nonce = 1;
			if (nonce == 255)
				nonce = 254;
			rand += nonce;
		}

		if (range != NULL) {
			uint32_t mask;

			addr_btom(range->addr_bits, &mask, IP_ADDR_LEN);
			mask = ntohl(mask);

			rand &= ~mask;
			rand |= ntohl(range->addr_ip) & mask;
			rand = htonl(rand);
			valid = 1;
		}

		addr_pack(&src, ADDR_TYPE_IP, IP_ADDR_BITS, &rand, IP_ADDR_LEN);
		if (!valid)
			valid = valid_ip(&src);

	}

	*result = src;
}

u_short
make_sport(int ttl, int count)
{
	u_short port = NTTL_BASE_SPORT;

	port += ttl * 10 + count;

	return (port);
}

u_short
make_dport(int ttl, int count)
{
	u_short port = NTTL_BASE_DPORT;

	port += ttl * 10 + count;

	return (port);
}

int
decode_sport(u_short sport, int *ttl, int *count)
{
	sport = ntohs(sport);

	sport -= NTTL_BASE_SPORT;

	/* XXX - Error checking */
	*count = sport % 10;
	*ttl = sport / 10;

	return (0);
}

/* Decode response packet for TTL scanning */

void
decode_tcp_probe(struct addr *addr, struct ip_hdr *ip, int type, int code)
{
	struct probe *probe;
	struct addr dst;
	struct tcp_hdr *tcp;
	u_short iphlen, iplen;
	int ttl, count;

	iphlen = ip->ip_hl << 2;
	if (iphlen < sizeof(struct ip_hdr))
		return;
	
	iplen = ntohs(ip->ip_len);
	if (iplen < iphlen + sizeof(struct tcp_hdr))
		return;

	tcp = (struct tcp_hdr *)((u_char *)ip + iphlen);
	if (type != DECODE_TCP_TCP) {
		addr_pack(&dst, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_dst, IP_ADDR_LEN);

		decode_sport(tcp->th_sport, &ttl, &count);
		
		/* Use TCP sequence number and IP id to detect scrubbers */
		/* Check if have a probe outstanding for this packet */
		if ((probe = find_probe(&dst)) == NULL)
			return;

		/*
		ttl = ip->ip_id & 255;
		if (ttl != (ip->ip_id / 256))
			return;
		*/
	} else {
		decode_sport(tcp->th_dport, &ttl, &count);
		
		if ((probe = find_probe(addr)) == NULL)
			return;
	}

	/* Store result and remove timeout */
	store_result(probe, addr, ttl, count, type, code);
}

int
tcp_probe(struct addr *saddr, struct addr *addr, int ttl, int count)
{
	u_char pkt[1500];
	struct tcp_hdr *tcp;
	u_int iplen;
	int res;

	tcp = (struct tcp_hdr *)(pkt + IP_HDR_LEN);
	tcp_pack_hdr(tcp, make_sport(ttl, count), 80, 
	    ntohl(addr->addr_ip), 0, 
	    TH_SYN, 0x8000, 0);

	iplen = IP_HDR_LEN + (tcp->th_off << 2);

	/* Src and Dst are reversed both for ip and tcp */
	ip_pack_hdr(pkt, 0, iplen, ttl * 256 + ttl,
	    IP_DF, ttl,
	    IP_PROTO_TCP, saddr->addr_ip, addr->addr_ip);

	ip_checksum(pkt, iplen);
	
	if ((res = ip_send(nttl_ip, pkt, iplen)) != iplen) {
		warn("%s: ip_send(%d): %s", __func__, res, addr_ntoa(addr));
		return (-1);
	}

	return (0);
}

void
decode_udp_probe(struct addr *addr, struct ip_hdr *ip, int type, int code)
{
	struct probe *probe;
	struct addr dst;
	struct udp_hdr *udp;
	u_short iphlen;
	int ttl, count;

	iphlen = ip->ip_hl << 2;
	if (iphlen < IP_HDR_LEN)
		return;

	udp = (struct udp_hdr *)((u_char *)ip + iphlen);
	addr_pack(&dst, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_dst, IP_ADDR_LEN);

	decode_sport(udp->uh_sport, &ttl, &count);
		
	/* Use TCP sequence number and IP id to detect scrubbers */
	/* Check if have a probe outstanding for this packet */
	if ((probe = find_probe(&dst)) == NULL)
		return;

	if (code == ICMP_UNREACH_PORT)
		type = DECODE_TCP_TCP; /* XXX - bad cludge to indicate end */

	/* Store result and remove timeout */
	store_result(probe, addr, ttl, count, type, code);
}

int
udp_probe(struct addr *saddr, struct addr *addr, int ttl, int count)
{
	u_char pkt[1500];
	struct udp_hdr *udp;
	struct udp_payload *payload;
	u_int iplen;
	int res;

	udp = (struct udp_hdr *)(pkt + IP_HDR_LEN);
	udp_pack_hdr(udp, make_sport(ttl, count), 32768 + ttl,
	    UDP_HDR_LEN + sizeof(struct udp_payload));

	payload = (struct udp_payload *)(pkt + IP_HDR_LEN + UDP_HDR_LEN);
	payload->ttl = ttl;
	payload->count = count;

	iplen = IP_HDR_LEN + UDP_HDR_LEN + sizeof(struct udp_payload);

	/* Src and Dst are reversed both for ip and udp */
	ip_pack_hdr(pkt, 0, iplen, ttl * 256 + ttl,
	    IP_DF, ttl,
	    IP_PROTO_UDP, saddr->addr_ip, addr->addr_ip);

	ip_checksum(pkt, iplen);
	
	if ((res = ip_send(nttl_ip, pkt, iplen)) != iplen) {
		warn("%s: ip_send(%d): %s", __func__, res, addr_ntoa(addr));
		return (-1);
	}

	return (0);
}

int
main(int argc, char *argv[])
{
	struct event sigterm_ev, sigint_ev, sighup_ev;
	struct addr daddr, saddr;
	int duration = 10;
	int c;
	char *dev = NULL;

	while ((c = getopt(argc, argv, "uD:r:ts:d:m:M:i:h?")) != -1) {
		switch (c) {
		case 'u':
			fnprobe = udp_probe;
			break;
		case 'm':
			nttl_maxprobes = atoi(optarg);
			if (nttl_maxprobes == 0)
				usage();
			break;
		case 'M':
			nttl_maxinflight = atoi(optarg);
			if (nttl_maxinflight == 0)
				usage();
			break;
		case 'i':
			dev = optarg;
			break;
		case 's':
			if (addr_aton(optarg, &saddr) == -1)
				errx(1, "Illegal network range: %s", optarg);
			nttl_srange = &saddr;
			break;
		case 'd':
			if (addr_aton(optarg, &daddr) == -1)
				errx(1, "Illegal network range: %s", optarg);
			nttl_range = &daddr;
			break;
	
		/* For TCP performance scanning */

		case 'D':
			duration = atoi(optarg);
			break;
		case 'r':
			tcprate = atoi(optarg);
			break;
		case 't':
			tcpmode = 1;
			break;
		case 'h':
			/* FALLTHROUGH */
		case '?':
			/* FALLTHROUGH */
		default:
			usage();
			/* not reached */
		}
	}

	argc -= optind;
	argv += optind;

	if (argc > 0) {
		fprintf(stderr, "Too many arguments.\n");
		usage();
	}

	if ((nttl_rand = rand_open()) == NULL)
		err(1, "rand_open");

	if ((nttl_ip = ip_open()) == NULL)
		err(1, "ip_open");

	/* Initalize libevent but without kqueue because of bpf */
	setenv("EVENT_NOKQUEUE", "yes", 0);
	event_init();

	interface_initialize();

	/* Initialize the specified interfaces */
	interface_init(dev, 0, NULL);

	signal_set(&sigint_ev, SIGINT, nttl_signal, NULL);
	signal_add(&sigint_ev, NULL);
	signal_set(&sigterm_ev, SIGTERM, nttl_signal, NULL);
	signal_add(&sigterm_ev, NULL);
	signal_set(&sighup_ev, SIGHUP, nttl_signal, NULL);
	signal_add(&sighup_ev, NULL);

	SPLAY_INIT(&probes);

	if (!tcpmode)
		add_probes(nttl_range);
	else {
		if (nttl_range == NULL || nttl_srange == NULL) {
			fprintf(stderr,
			    "Need both source and destination range.\n");
			usage();
		}
		tcp_setup(duration);
	}
	
	event_dispatch();

	syslog(LOG_ERR, "Kqueue does not recognize bpf filedescriptor.");

	return (0);
}
