/*
 * Copyright 2002, 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_IOCCOM_H
#include <sys/ioccom.h>
#endif
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/tree.h>
#include <sys/wait.h>
#include <sys/queue.h>

#include <pcap.h>

#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#include <unistd.h>
#include <getopt.h>
#include <dnet.h>

#undef timeout_pending
#undef timeout_initialized

#include <event.h>

#include "nttlscan.h"
#include "interface.h"

extern struct addr *nttl_range;		/* destination ip addresses */
extern struct addr *nttl_srange;	/* source ip addresses */
extern struct interface *nttl_inter;
extern rand_t *nttl_rand;
extern ip_t   *nttl_ip;

int tcprate = 100;
static struct timeval tcpstart;
static struct timeval tcpstop;
static struct event ev_send;
static struct event ev_ende;

/* Statistics for tcp mode */
struct {
	int nsent;			/* number of probes sent */
	int nsuccess;			/* number of successful exchanges */
	int nreset;			/* number of connection resets */
	int nsynack;			/* number of syn|acks */
} tcpstats;

/* Decode response packet if we are in TCP scan mode */

void
decode_tcp_connect(struct addr *addr, struct ip_hdr *ip)
{
	struct addr dst;
	struct tcp_hdr *tcp;
	u_short iphlen, iplen;

	iphlen = ip->ip_hl << 2;
	if (iphlen < sizeof(struct ip_hdr))
		return;
	
	iplen = ntohs(ip->ip_len);
	if (iplen < sizeof(struct ip_hdr) + sizeof(struct tcp_hdr))
		return;

	/* Check address ranges */
	addr_pack(&dst, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_dst, IP_ADDR_LEN);
	if (!addr_inrange(nttl_range, addr) ||
	    !addr_inrange(nttl_srange, &dst))
		return;

	tcp = (struct tcp_hdr *)((u_char *)ip + iphlen);
	if (tcp->th_flags & TH_RST) {
		tcpstats.nreset++;
		return;
	} else if (tcp->th_flags == (TH_SYN|TH_ACK)) {
		u_char pkt[1500];
		struct tcp_hdr *res_tcp;
		u_int iplen;
		int res;

		tcpstats.nsynack++;

		/* 
		 * Send back ACK and then send data packet.
		 */
		res_tcp = (struct tcp_hdr *)(pkt + IP_HDR_LEN);
		tcp_pack_hdr(res_tcp,
		    ntohs(tcp->th_dport), 80, 
		    ntohl(tcp->th_ack), ntohl(tcp->th_seq) + 1, 
		    TH_ACK, 0x8000, 0);

		iplen = IP_HDR_LEN + (res_tcp->th_off << 2);

		/* Src and Dst are reversed both for ip and tcp */
		ip_pack_hdr(pkt, 0, iplen, rand_uint16(nttl_rand),
		    IP_DF, 128,
		    IP_PROTO_TCP, dst.addr_ip, addr->addr_ip);
		
		ip_checksum(pkt, iplen);
		
		if ((res = ip_send(nttl_ip, pkt, iplen)) != iplen)
			err(1, "%s: ip_send(%d): %s",
			    __func__, res, addr_ntoa(addr));

		/*
		 * Send back data packet
		 */
		res_tcp = (struct tcp_hdr *)(pkt + IP_HDR_LEN);
		tcp_pack_hdr(res_tcp,
		    ntohs(tcp->th_dport), 80, 
		    ntohl(tcp->th_ack), ntohl(tcp->th_seq) + 1, 
		    TH_ACK, 0x8000, 0);

		iplen = IP_HDR_LEN + (res_tcp->th_off << 2) + 
		    sizeof(TCP_CONTENT);

		/* Src and Dst are reversed both for ip and tcp */
		ip_pack_hdr(pkt, 0, iplen, rand_uint16(nttl_rand),
		    IP_DF, 128,
		    IP_PROTO_TCP, dst.addr_ip, addr->addr_ip);

		memcpy(pkt + IP_HDR_LEN + (res_tcp->th_off << 2),
		    TCP_CONTENT, sizeof(TCP_CONTENT));

		ip_checksum(pkt, iplen);
		
		if ((res = ip_send(nttl_ip, pkt, iplen)) != iplen)
			err(1, "%s: ip_send(%d): %s",
			    __func__, res, addr_ntoa(addr));
	} else if (tcp->th_flags == TH_ACK) {
		u_char pkt[1500];
		struct tcp_hdr *res_tcp;
		u_int iplen;
		int res, ack = 0;

		/* This assumes an echo server */
		iplen = ntohs(ip->ip_len) - IP_HDR_LEN - (tcp->th_off << 2);
		if (iplen < sizeof(TCP_CONTENT)) {
			ack = 1;
		} else {
			u_char *p = (u_char *)ip;
			int off = IP_HDR_LEN + (tcp->th_off << 2);

			off += iplen - sizeof(TCP_CONTENT);

			/* We might get our reply with other data */
			if (memcmp(p + off, TCP_CONTENT, sizeof(TCP_CONTENT)))
				ack = 1;
			else
				tcpstats.nsuccess++;
		}

		/* 
		 * Tear down connection or ACK the data
		 */
		res_tcp = (struct tcp_hdr *)(pkt + IP_HDR_LEN);
		tcp_pack_hdr(res_tcp,
		    ntohs(tcp->th_dport), 80, 
		    ntohl(tcp->th_ack), ntohl(tcp->th_seq) + 1, 
		    ack ? TH_ACK : TH_RST, 0x8000, 0);

		iplen = IP_HDR_LEN + (res_tcp->th_off << 2);

		/* Src and Dst are reversed both for ip and tcp */
		ip_pack_hdr(pkt, 0, iplen, rand_uint16(nttl_rand),
		    IP_DF, 128,
		    IP_PROTO_TCP, dst.addr_ip, addr->addr_ip);
		
		ip_checksum(pkt, iplen);
		
		if ((res = ip_send(nttl_ip, pkt, iplen)) != iplen)
			err(1, "%s: ip_send(%d): %s",
			    __func__, res, addr_ntoa(addr));
	}

}

void
tcp_connect_send(int fd, short what, void *arg)
{
	struct event *ev = arg;
	struct timeval tv;
	int ntotal, nprobes, i;

	/* Schedule the next probe */
	timerclear(&tv);
	tv.tv_usec = 1000000L / tcprate;
	evtimer_add(ev, &tv);

	gettimeofday(&tv, NULL);
	timersub(&tv, &tcpstart, &tv);

	ntotal = tv.tv_sec * tcprate + (tv.tv_usec * tcprate) / 1000000L;
	nprobes = ntotal - tcpstats.nsent;

	for (i = 0; i < nprobes; i++) {
		struct addr src, dst;

		make_random_ip(nttl_srange, &src);
		make_random_ip(nttl_range, &dst);
		/* XXX - Send the probe here */
		tcp_probe(&src, &dst, 128, i);
		tcpstats.nsent++;
	}
}


void
tcp_setup_end2(int fd, short what, void *arg)
{
	struct pcap_stat stat;
	float rate, rret;

	rate = (float)tcpstats.nsent /(tcpstop.tv_sec + (float)tcpstop.tv_usec/1000000L);
	rret = (float)tcpstats.nsuccess / (float)tcpstats.nsent;

	fprintf(stdout, "%f: %f\n", rate, rret);
	
	if (pcap_stats(nttl_inter->if_pcap, &stat) < 0) {
		fprintf(stderr, "pcap_stats: %s\n",
		    pcap_geterr(nttl_inter->if_pcap));
	} else {
		fprintf(stderr, "%d packets received by filter\n",
		    stat.ps_recv);
		fprintf(stderr, "%d packets dropped by kernel\n",
		    stat.ps_drop);
	}

	exit(0);
}

void
tcp_setup_end(int fd, short what, void *arg)
{
	struct timeval tv;

	gettimeofday(&tcpstop, NULL);
	timersub(&tcpstop, &tcpstart, &tcpstop);

	event_del(&ev_send);
	fprintf(stderr, "Stopped sending, waiting for replies ...\n");

	evtimer_set(&ev_ende, tcp_setup_end2, &ev_ende);
	timerclear(&tv);
	tv.tv_sec = 10;
	evtimer_add(&ev_ende, &tv);
}

void
tcp_setup(int duration)
{
	struct timeval tv;

	gettimeofday(&tcpstart, NULL);

	evtimer_set(&ev_send, tcp_connect_send, &ev_send);
	timerclear(&tv);
	tv.tv_usec = 1000000L / tcprate;
	evtimer_add(&ev_send, &tv);

	evtimer_set(&ev_ende, tcp_setup_end, &ev_ende);
	timerclear(&tv);
	tv.tv_sec = duration;
	evtimer_add(&ev_ende, &tv);
}
